/**
 * Represent a PDF File with some
 * utils to manipulate it.
 *
 * A least we need :
 *  - Create an empty file
 *  - Read an empty file
 *  - Add a page
 *  - Remove a page
 *  - Add text
 *  - Remove text
 *  - Add an image
 *  - remove an image
 *  - Extract all text from a PDF
 *  - Extract all images from a PDF
 *  - Make a thumbnail for a given page
 *
 *  - More Difficult : Read pdf produce by oO
 *
 *
 * @author vingtetun
 */

const NOXREF = "The document doesn't contains a startxref section"


/**
 * A PDF is divided into 4 parts
 *
 *  +---------------+
 *  |     HEADER    |
 *  +---------------+
 *  |     BODY      |
 *  |               |
 *  +---------------+
 *  |     XREF      |
 *  +---------------+
 *  |     TRAILER   |
 *  +---------------+
 *
 *
 *  Header :
 *     Identify a document as a PDF document via a string "%PDF-1.1".
 *
 *  Body :
 *    Collection of objects organized as a tree.
 *    Each object has three essential components: 
 *      - A number
 *      - A fixed position in the PDF file (an offset)
 *      - A content. 
 *
 *  Xref :
 *    Where to find each object (offset)
 *
 *  Trailer :
 *    Where to find XREF
 *    Contains the root object (start of the tree)
 *    Contains the info object (meta information)
 */

const IOService = Components.classes["@mozilla.org/network/io-service;1"]
                            .getService(Components.interfaces.nsIIOService);
const LocalFile = Components.Constructor("@mozilla.org/file/local;1","nsILocalFile","initWithPath");
const BinaryInputStream = Components.Constructor("@mozilla.org/binaryinputstream;1","nsIBinaryInputStream", "setInputStream");
const FileInputStream = Components.Constructor("@mozilla.org/network/file-input-stream;1","nsIFileInputStream","init");
const StringInputStream = Components.Constructor("@mozilla.org/io/string-input-stream;1","nsIStringInputStream");
const ScriptableInputStream = Components.Constructor("@mozilla.org/scriptableinputstream;1","nsIScriptableInputStream", "init");


/**
 * Firstly we read the full document, find the xref offset
 * (indicate by startxref ) and get all the object block.
 * Each block are passed to PDFObject to be compute as needed.
 *
 * XXX In a first step we try to only recover the EXACT text of
 * the PDF document and the pages count.
 * Once we have this we can go further.
 *  - Do an html page - With xslt?
 *  - Do a screenshot of the first page to have a preview of
 *    a pdf document!
 *
 * Number of lines of a PDF Document are summarized in a Pages object
 * with no parent.
 * There can be a lot of pages object but subPages have a \Parent
 * line which refer to the parent Pages.
 *
 */
var PDFDocument = function(aFilePath) {
  //Constructor
  var pdfStream = new FileInputStream(new LocalFile(aFilePath), 0x01, 0444, null);
  pdfStream.QueryInterface(Components.interfaces.nsILineInputStream);
  pdfStream.QueryInterface(Components.interfaces.nsISeekableStream);

 /* 
 * Specification says the startxref should live in the last
 * 1024 bytes of the data.
 *
 * If we don't found it in here, the doc is probably not a PDF
 * or a malformed one!!!
 */
  var docSize = pdfStream.available();
  pdfStream.seek(pdfStream.NS_SEEK_END, (docSize < 1025) ? -docSize + 1
                                                         : -1025);
  var startxrefIsFound = false;
  var streamNotFinish = true;
  this.startxref = 0;
  while (!startxrefIsFound && streamNotFinish) {
    let line = {};
    streamNotFinish = pdfStream.readLine(line);
    if (line.value == "startxref") {
      startxrefIsFound = true;
      pdfStream.readLine(line);
      this.startxref = parseInt(line.value,10);
    };
  }

  if (this.startxref == 0) //there is an error
    throw new Error("Parsing PDF Document error : " + NOXREF);

  this.xref = new XREF(pdfStream, this.startxref);
  
  // Search for page tree once we have the catalog offset
  var catalog = Tokenizer.exec(pdfStream, this.xref.catalogOffset);
  for(let i = 1; i < catalog.length; i++){
    switch (catalog[i]) {
      case "Pages":
        pagesOffset = this.xref.dictionnary[catalog[i+1]].offset;
        break;
    }
  }

  var pages = Tokenizer.exec(pdfStream, pagesOffset);
  var pageKids = [];
  for (let i = 1; i < pages.length; i++) {
    switch (pages[i]) {
      case "Count":
        var pageCount = parseInt(pages[i+1], 10);
        break;

      case "Kids":
        var header = pages[++i];
        for (let j=0; j < header.length; j += 3) {
          //XXX Should do a tree build here,not a so simple one
          pageKids.push(this.xref.dictionnary[parseInt(header[j],10)].offset);
        }
        break;
    }
  }
    
  dump('Nombre de page:' + pageCount + '\n');

  // Build data
  if (pageKids.length != 0) {
    var pagesContent = [];

    for (let i = 0; i < pageKids.length; i++) {
      let contentsOffset = [];

      let data = Tokenizer.exec(pdfStream, pageKids[i]);
      for (let i = 1; i < data.length; i++) {
        switch (data[i]) {
          case "Contents":
            var header = data[++i];
            for (let j=0; j < header.length; j += 3) {
              contentsOffset.push(this.xref.dictionnary[parseInt(header[j],10)].offset);
            }
            let internalData = [];
            for (let j = 0; j < contentsOffset.length; j++) {
              //internalData.push(pdfutils.readPageObject(pdfStream, parseInt(contentsOffset[j], 10)));
              internalData.push(Tokenizer._execPage(pdfStream, parseInt(contentsOffset[j],10)));
            }
            pagesContent.push(internalData);
            break;
        }
      }
    }

    //Finally dump the data...
    for (let i=0; i < pagesContent.length; i++) {
      let text = "";

      let pageContent = pagesContent[i];
      for (let j=0; j < pageContent.length; j++) {
        //Cause of a bug in ff, it seems we can use nsIStreamConverterService.convert
        //directly.
        //So the workaround is to use a nsIInputStreamPump and make a call to 
        //asynConvertData throught the nsIStreamConverterService.
        //Can find any bug on bugzilla about that
        var [position, [type, length]] = pageContent[j];
        if (type=="adress") {
          var adress = Tokenizer.exec(pdfStream, this.xref.dictionnary[length].offset);
          length = parseInt(adress[3], 10);
        }

        if (length) {
          ConverterService.convert(pdfStream, position, length, i, callback);
        }
      }
    }
  }
  this.pages = 0;
  //pdfStream.close();
}

