/**
 * This is a simple PDF Tokenizer to
 * handle content and the different
 * object in it.
 *
 * I feel that the most complicate task here are:
 *  - Handle the last data (xref, trailer)
 *  - Don't sucks on stream...endstream
 *
 *
 *  In a classical language analysis process
 *  we must do some kind of parsing, but here we
 *  suppose the PDF is right so we skip
 *  this phase and reduce our analysis to 
 *  building a lexical analysis and a 
 *  semantic one.
 */




/**
 * @name Tokenizer
 * @static
 * Tokenize a stream, reading from the
 * beginning of an object to the end of it.
 * The end depend on the type of the object:
 *  - Trailer : %EOF
 *  - Obj : endobj
 */
var Tokenizer = {};


/**
 * @name Tokenize.exec
 * @static
 * Return an array of tokens.
 *
 * @param {nsIInputStream} aStream - The stream to tokenize from
 * @param {int} anOffset - The offset to begin the tokenization
 */
Tokenizer.exec = function(aStream, anOffset) {
  aStream.seek(aStream.NS_SEEK_SET, anOffset);
  var tokens = [];
  var data = "";
  while (!this._isEnd(tokens)) {
    let line = {};
    aStream.readLine(line);
    data += line.value + "\n";
    tokens = tokens.concat(this._exec(line.value));
  }
  PDFParser.syntaxic.exec( data );
  return Tokenizer._rebuild(tokens);
};

Tokenizer.old_exec = function(aStream, anOffset) {
  aStream.seek(aStream.NS_SEEK_SET, anOffset);
  var tokens = [];
  while (!this._isEnd(tokens)) {
    let line = {};
    aStream.readLine(line);
    tokens = tokens.concat(this._exec(line.value));
  }
  return Tokenizer._rebuild(tokens);
};

/**
 * @name Tokenizer._rebuild
 * @static
 * @private
 * The postscript notation use []
 * to keep some array of data, futhermore
 * stream in pdf contains data enclose in ()
 * So we try to keep these stuff intact to
 * facilitate the work of the rest
 * @param {array} anArray - The array to rebuild
 */
Tokenizer._rebuild = function(anArray) {
  //Clean empty token
  var del = 0;
  for (let i = 0; i < anArray.length; i++) {
    if (!anArray[i] || anArray[i].length == 0) {
      anArray.splice(i, 1);
    }
  }

  var rebuilded = [];
  var _lock = false;
  var _temp = [];
  for (let i = 0; i < anArray.length; i++) {
    if (anArray[i] == "[")
      _lock = true;
    else {
      if (_lock && anArray[i] != "]")
        _temp.push(anArray[i]);
      else {
        if (anArray[i]=="]") {
          _lock = false;
          rebuilded.push(_temp);
          _temp = [];
        }
        else
          rebuilded.push(anArray[i]);
      }
    }
  }
  return rebuilded;
};


/**
 * @name Tokenizer._isEnd
 * @static
 * @private
 * Say if a token if the end or not
 * @param {array} anArray - The token to check for
 * @return {boolean} True if the tokens is a end tokens
 */
Tokenizer._isEnd = function(anArray) {
  if (!anArray.length)
    return false;

  switch (anArray[anArray.length-1]) {
    case "endobj":
    case "startxref":
      return true;
      break;

    default:
      break;
  }

  return false;
};


/**
 * @name Tokenizer._execPage
 * Parse a stream containing a page,
 * and replacing [stream] with an array
 * containing [offset, length].
 * Return at the end an array of tokens
 *
 * @param {nsIInputStream} aStream - The stream to tokenize from
 * @param {int} anOffset - The offset to begin the tokenization
 * @return {array} Tokens array
 *
 */
Tokenizer._execPage = function(aStream, anOffset) {
  aStream.seek(aStream.NS_SEEK_SET, anOffset);
  var toto = new BinaryInputStream(aStream);

  var previousWords = "";
  while (/stream\cM?\s$/.test(previousWords) == false) {
    previousWords += toto.readBytes(1);
  }

  // ok. now tokenize the gotten stuff
  var tokens = this._execString(previousWords);
  var size = 0;
  for (let i = 0; i < tokens.length; i++) {
    switch (tokens[i]) {
      case "Length":
        if (/R\cM?/.test(tokens[i+3])) {
          //arghh... search in an other object
          size = ['adress', parseInt(tokens[i+1], 10)];
          //XXX breaking stuff!!!!
          anOffset +=1;
        }
        else{
          size = ['integer', parseInt(tokens[i+1], 10)];
        }
      break;
    }
  }

  return [anOffset + previousWords.length , size]
};


/**
 * @name Tokenizer._exec
 * @static
 * @private
 * Return an array of tokens.
 *
 * @param {string} aString - The string to tokenize
 * @return {array} The tokens
 */
Tokenizer._exec = function(aString) {
  // XXX Do a better regexp here
  
  aString = aString.replace(/\cM/,"");
  aString = aString.replace(/([a-z\d])([\[\]])/,"$1 $2");
  aString = aString.replace(/[^\n]endstream/g,"\nendstream");
  if (aString.indexOf("[") != -1) {
    aString = aString.replace(/\[([0-9])/,"[ $1");
  }
  if (aString.indexOf("]") != -1) {
    aString = aString.replace(/([a-z0-9])\]/i,"$1 ]");
  }
  var separator = /\cM?[\s\/]+/gi;
  return aString.split(separator);
};



/**
 * @name @Tokenizer._execString
 * Tokenize a string and performs some
 * (un)necessary operations depending on
 * the semantic aspect of the tokens.
 *
 * All data array are returned as a unique string
 * of precompute data.
 * For examples [<54>78<65>11<73>6<74>3<2050>6<44>-3<46>56<2E2E2E>]
 * will be returned as "Test PDF..."
 *
 * This way the implentation of command handling
 * is really simpler.
 *
 * @param {string} The string to handle
 * @return {array} anArray of data
 *
 */

Tokenizer._execString = function(aString) {
  /*aString = aString.replace(/<([a-f\d])/gi, "( $1");
  aString = aString.replace(/([a-f\d])>/gi, "$1 )");*/

  if (aString.indexOf("(") != -1) {
    aString = aString.replace(/([^\\])\(/gi,"( $1");
  }
  if (aString.indexOf(")") != -1) {
    aString = aString.replace(/([^\\ ])\)/gi,"$1 )");
  }
  aString = aString.replace(/(\[)/g," ( ");
  aString = aString.replace(/(\])/g," ) ");

  //Search for T[J|j] not enclosed
  var search = /T[J|j][ \r\n]/g;
  var replaceStrings = [];
  while ((rv = search.exec(aString)) != null) {
    // We have found some, rebuild the string
    var start = search.lastIndex;
    var isEnd = false;
    var isStart = false;
    var endChar = "";
    var position = 1;
    var string = [];
    var car = "";
    while (!isStart) {
      position++;
      car = aString.substring(start-position, start - position+1);
      if(/[\]\)\>]/.test(car)){
        isStart = true;
        switch(car){
          case "]":
            endChar = /\[/;
          break;
          case ")":
            endChar = /\(/;
          break;
          case ">":
            endChar = /\</;
          break;
        }
      }
    }
    while (!isEnd) {
      position++;
      car = aString.substring(start-position, start - position + 1);
      if (endChar.test(car)) {
        isEnd = true;
      }
      else
        string.push(car);
    }
    replaceStrings.push(string.reverse());
  }
  //XXX replace the begin/end by aconvenient stuff

  //Performs a clean up of these stuffs
  function hexToAscii(anHex) {
    if (anHex.length % 2 != 0)
      anHex += 0;

    anHex = anHex.replace(/([\da-f]{2})/gi,"%$1");
    return unescape(anHex);
  }

  var toDecode = "";
  var decode = false;
  for each (item in replaceStrings) {
    let finalString = "";
    for(var j = 0; j < item.length; j++) {
      if (item[j] == "<") {
        decode = true;
      }
      else if (item[j] == ">") {
        decode = false;
        finalString += hexToAscii(toDecode);
        toDecode = "";
      }
      else if (decode) {
        toDecode += item[j];
      }
    }

    if (finalString != "")
      aString = aString.replace(item.join(""), " " + finalString + " ");
  }

  var separator = /\r?[\s\/]+/gi;
    
  return this._rebuildString(aString.split(separator));
};


Tokenizer._rebuildString = function(anArray) {
  var rebuilded = [];
  var _lock = false;
  var _temp = [];
  for (let i = 0; i < anArray.length; i++) {
    if (anArray[i] == "(" && _lock == false) {
      _lock = true;
    }
    else {
      if (_lock && anArray[i] != ")")
        _temp.push(anArray[i]);
      else{
        if (anArray[i] == ")") {
          _lock = false;
          rebuilded.push(_temp);
          _temp = [];
        }
        else
          if (anArray[i])
            rebuilded.push(anArray[i]);
      }
    }
  }
  //clean up \\
  /*for(let i = 0 ; i < rebuilded.length ; i++){
    if(rebuilded[i].indexOf("(") != -1 || rebuilded[i].indexOf(")")){
      if(rebuilded[i].replace)
        rebuilded[i].replace(/\\([\(\)])/gi,"$1");
    }
  }*/
  return rebuilded;
};

